# --------------------------------------------------------------------
# cookbook.py
#
# Author: Lain Musgrove (lain.proliant@gmail.com)
# Date: Tuesday April 4, 2023
#
# Distributed under terms of the MIT license.
#
# NOTE: This module is deprecated.  The functions and classes that used
# to be in this module have been moved to other modules.  Please do not
# add further functions or classes to this module.  The ones imported
# below are included for backwards compatibility with older builds.
#
# This file will be removed in the next major version (v7.x).
# --------------------------------------------------------------------

from xeno.recipe import recipe
from xeno.recipes.shell import ShellRecipe, sh
